/*
Module : DDXFolder.cpp
Purpose: Implementation for a MFC DDX_ routine to get a directory/folder using the IFileDialog interface
Created: PJN / 26-03-1997
History: PJN / 16-07-1997 Added support for setting the initial directory when using SHBrowseForFolder
         PJN / 26-11-1997 Update to allow the folder selection dialogs to be used withou the DDX stuff
         PJN / 16-09-1998 1. Updated all the documentation to be HTML based as with my other shareware / freeware.
                          2. Unicode enabled all the code and provision of Unicode build configurations.
                          3. VC 5 mak files provided now as standard
                          4. Option to have non read only edit control.
                          5. Provision of a DDV function
                          6. General tidy up of the sample app including removing all the AppWizard
                             generated comments. 
                          7. All code now compiles cleanly at warning level 4
                          8. Replaced all TRACE0 calls with TRACE
                          9. Changed name of main function from DDX_GetFolder to DDX_FolderControl
                          10. Module name has been changed from ddxgetfolder to ddxfolder.
                          11. Sample app now allows read only state of widget to be toggled
                          13. Addition of a DDX_FolderValue function
                          14. Change the ID's of the strings in the resource table which the code uses
         PJN / 17-09-1998 1. Minor update to some comments and the documentation
         PJN / 11-12-1999 1. Cosmetic item where now a 1 pixel space is included between the edit control and 
                          the "..." button
         PJN / 14-05-2000 The "..." button now supports being tabbed into i.e. the WS_TABSTOP style
         PJN / 17-05-2000 The width of the underlying edit control is now shrunk when the "..." button is created.
                          This makes layout of the control using the resource designer a lot simpler.
         PJN / 18-07-2000 1. Removed an unreferenced variable from the demo app module "testgetfolderDlg.cpp"
                          2. Width of the edit button is now determined from the text which it displays.
         PJN / 24-08-2001 1. Fixed a problem with the Z-Order of the control being set incorrectly
                          2. Updated copyright message
         PJN / 09-01-2003 1. Updated copyright details.
                          2. Changed a class name from "CModifyButton" to the better named "CDDXFileModifyButton".
                          3. Enabled state of control is now reflected to buddy button edit control. Thanks to 
                          "matro" for pointing out this problem.
                          4. Made the "CGetFolderControl::Edit" virtual for easier customisation.
                          5. Made all strings used by the classes easier to modify by making them virtual
                          methods of the class
                          6. Made the "CGetFolderControl::AddEditButton" method virtual for easier customisation.
         PJN / 19-07-2006 1. Updated copyright details.
                          2. Addition of a CDDXFOLDER_EXT_CLASS preprocessor macro to allow the code to be
                          easily added to an extension dll.
                          3. Optimized CDDXFolderModifyButton constructor code
                          4. Removed support for the old style directory picker dialog which used a modified
                          version of the File Open dialog.
                          5. Changed the names of the constants used for DDV_FolderControl
                          6. Added support for a DDXGETFOLDER_FOLDERMUSTEXIST flag
                          7. Removed the GetFolder and SetFolder methods as they are not really required
                          8. Verified code clean compiles in VC2005
         PJN / 29-07-2006 1. Initial enabled state of the edit control is now reflected into the edit button
                          when AddEditButton is called. 
         PJN / 11-08-2006 1. Reworked all the use of string resources. Now all the string resources used
                          by the control are all loaded in one single virtual function. In addition if you 
                          do not want to use inheritance, all the strings can be customized via Get/Set 
                          methods.
                          2. Broke the DDV functionality into seperate discreet DDV functions
                          3. The dialog control id of the edit button can now be specified via the SubclassEdit
                          and DDX_FilenameControl functions
         PJN / 19-08-2006 1. Control now takes advantage of the AutoCompletion functionality (IAutoComplete et al) 
                          built into Windows to improve the UI experience. If enabled then a drop down list
                          will appear underneath the edit control which allows the user to quickly select a specific
                          folder. The code uses the COM interfaces (IAutoComplete) to AutoComplete instead of the 
                          more simplified functionality provided by SHAutoComplete. This allows the code to customize
                          the auto completion UI so that only file system folders are selectable.
         PJN / 25-08-2006 1. Minor update to the name of one of the string resources used by DDXFolder
         PJN / 19-11-2007 1. Minor update to display HRESULT's correctly.
                          2. Minor tidy up to remove old style AppWizard comments
         PJN / 21-11-2007 1. The code now uses the value 32 instead of the enum ACLO_FILESYSDIRS in the call to 
                          IID_IACList2::SetOptions. This is because newer versions of the Windows SDK make this value 
                          dependent on the setting of the _WIN32_IE preprocessor macros. Thanks to Alexey Kuznetsov
                          for reporting this issue.
         PJN / 18-12-2008 1. Updated copyright details
                          2. InitializeAutoComplete now returns a HRESULT instead of a BOOL
                          3. DDV_FolderControlNotFile function has been renamed to DDV_FolderControlIsFolder
                          4. Removed VC 6 style AppWizard comments from the code
                          5. The code has now been updated to support VC 2005 or later only
                          6. Code now compiles cleanly using Code Analysis (/analyze)
                          7. Updated DDX_FolderControl code to be consistent with DDX_Control in VC 2008
         PJN / 10-09-2011 1. Updated copyright details.
                          2. Verified code compiles cleanly on VC 2010
                          3. Fixed an access violation bug in CDDXFolderModifyButton::PreTranslateMessage if the 
                          IDS_DDXFOLDER_EDIT_TEXT string resource is not loaded    
         PJN / 02-03-2014 1. Updated copyright details.
                          2. Updated MFC message maps in sample app and in DDXFile module to use C++ standard 
                          for pointers to member methods.
                          3. Reworked DDX_FolderControl to use DDX_Control for most of its functionality
                          4. Updated the code to clean compile on VC 2013 & 2012.
                          5. Updated the code to clean compile using /analyze
         PJN / 23-12-2014 1. If the main edit control is now hidden, the buddy buttton is now also hidden. The 
                          corrolary of if the main edit control is now shown, then the buddy button is now also
                          shown. Thanks to Rene Laviolette for requesting this update.
         PJN / 12-11-2015 1. Verified the code compiles cleanly on VC 2015.
                          2. Updated the sample apps main icon.
                          3. Updated copyright details.
         PJN / 23-09-2017 1. Updated copyright details.
                          2. Replaced CString::operator LPC*STR() calls throughout the codebase with 
                          CString::GetString calls
                          3. Replaced NULL throughout the codebase with nullptr. This means that the minimum 
                          requirement for the framework is now VC 2010.
                          4. Reworked the code in CGetFolderControl::InitializeAutoComplete to use 
                          ATL::CComPtr::CoCreateInstance.
                          5. All the code has had SAL annotations added
         PJN / 08-07-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
         PJN / 19-05-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 14-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 03-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 02-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 1997 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "DDXFolder.h"
#include "ChooseDirDlg.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CDDXFolderModifyButton, CButton) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
  ON_CONTROL_REFLECT(BN_CLICKED, &CDDXFolderModifyButton::OnClicked)
END_MESSAGE_MAP()

CDDXFolderModifyButton::CDDXFolderModifyButton() noexcept : m_pBuddy{nullptr},
                                                            m_bFirstPreTranslateMessageCall{true}
{
}

void CDDXFolderModifyButton::SetBuddy(_In_opt_ CGetFolderControl* pBuddy) noexcept
{
  m_pBuddy = pBuddy;
}

BOOL CDDXFolderModifyButton::PreTranslateMessage(MSG* pMsg)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pBuddy != nullptr);

  //create the tooltip
  if (m_bFirstPreTranslateMessageCall)
  {
    ASSERT(m_pBuddy != nullptr);
    CString sTTText{m_pBuddy->GetModifyButtonToolTipText()};
    if (sTTText.GetLength())
    {
      m_ToolTip.Create(this);
      m_ToolTip.Activate(TRUE);
      m_ToolTip.AddTool(this, sTTText);
    }
    m_bFirstPreTranslateMessageCall = false;
  }

  //give the tooltip a chance to handle the message if is is valid
  if (m_ToolTip.GetSafeHwnd() != nullptr)
    m_ToolTip.RelayEvent(pMsg);

  //Let the base class do its thing
  return __super::PreTranslateMessage(pMsg);
}

void CDDXFolderModifyButton::OnClicked()
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pBuddy != nullptr);

  //Ask the buddy to show the dialog
  m_pBuddy->Edit();
}


#pragma warning(suppress: 26440 26433)
BEGIN_MESSAGE_MAP(CGetFolderControl, CEdit) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_ENABLE()
  ON_WM_SHOWWINDOW()
END_MESSAGE_MAP()

CGetFolderControl::CGetFolderControl() noexcept : m_bAutoComplete{true},
                                                  m_dwAutoCompleteFlags{ACO_AUTOSUGGEST | ACO_FILTERPREFIXES}
{
}

bool CGetFolderControl::SubclassEdit(_In_ HWND hEdit, _In_ UINT nModifyButtonID)
{
  //test our inputs
  if (!IsWindow(hEdit))
  {
    ASSERT(false);
    TRACE(_T("CGetFolderControl::SubclassEdit, window handle is invalid!\n"));
    return false;
  }

  //subclass the control
  if (SubclassWindow(hEdit))
  {
    LoadStringResources();
    InitializeAutoComplete(); //Failing to setup auto complete is not considered a critical error
    return AddModifyButton(nModifyButtonID);
  }
  else
  {
    TRACE(_T("CGetFolderControl::SubclassEdit, Could not subclass edit control!\n"));
    ASSERT(false);
    return false;
  }
}

void CGetFolderControl::SetAutoComplete(_In_ bool bAutoComplete)
{
  //Save the setting in the member variable
  m_bAutoComplete = bAutoComplete;

  //Also let the auto complete object know
  if (m_pAutoCompleteObject != nullptr)
  {
    if (m_bAutoComplete)
    {
      const HRESULT hr{m_pAutoCompleteObject->SetOptions(m_dwAutoCompleteFlags)};
      if (FAILED(hr))
        TRACE(_T("CGetFolderControl::SetAutoComplete, failed to set AutoComplete object options, Error:%08X\n"), hr);
    }
    else
    {
      const HRESULT hr{m_pAutoCompleteObject->SetOptions(ACO_NONE)};
      if (FAILED(hr))
        TRACE(_T("CGetFolderControl::SetAutoComplete, failed to set AutoComplete object options, Error:%08X\n"), hr);
    }
  }
}

HRESULT CGetFolderControl::InitializeAutoComplete()
{
  if (m_bAutoComplete)
  {
    //Create the auto complete object
    HRESULT hr{m_pAutoCompleteObject.CoCreateInstance(CLSID_AutoComplete, nullptr, CLSCTX_INPROC_SERVER)};
    if (FAILED(hr))
    {
      TRACE(_T("CGetFolderControl::InitializeAutoComplete, failed to create AutoComplete object, Error:%08X\n"), hr);
      return hr;
    }

    //Create the auto complete source object
    ASSERT(m_pAutoCompleteSourceObject == nullptr);
    hr = m_pAutoCompleteSourceObject.CoCreateInstance(CLSID_ACListISF, nullptr, CLSCTX_INPROC_SERVER);
    if (FAILED(hr))
    {
      TRACE(_T("CGetFolderControl::InitializeAutoComplete, failed to create AutoComplete filesystem source object, Error:%08X\n"), hr);
      return hr;
    }

    //Tell the auto complete source object to only return file system directories
    hr = m_pAutoCompleteSourceObject->SetOptions(ACLO_FILESYSDIRS);
    if (FAILED(hr))
    {
      TRACE(_T("CGetFolderControl::InitializeAutoComplete, failed to set AutoComplete filesystem source object options, Error:%08X\n"), hr);
      return hr;
    }

    //Finally set the options and hook up the edit control
    hr = m_pAutoCompleteObject->SetOptions(m_dwAutoCompleteFlags);
    if (FAILED(hr))
    {
      TRACE(_T("CGetFolderControl::InitializeAutoComplete, failed to set AutoComplete object options, Error:%08X\n"), hr);
      return hr;
    }
    hr = m_pAutoCompleteObject->Init(GetSafeHwnd(), m_pAutoCompleteSourceObject, nullptr, nullptr);
    if (FAILED(hr))
    {
      TRACE(_T("CGetFolderControl::InitializeAutoComplete, failed to initialize AutoComplete object, Error:%08X\n"), hr);
      return hr;
    }
  }

  return S_OK;
}

bool CGetFolderControl::AddModifyButton(_In_ UINT nModifyButtonID)
{
  //Validate our parameters
  CWnd* pParent{GetParent()};
#pragma warning(suppress: 26496)
  AFXASSUME(pParent != nullptr);

  //Work out the dimensions of the modify button and the resized edit control
  CRect editRect;
  GetWindowRect(editRect);
  pParent->ScreenToClient(editRect);
  CRect btnRect;
  btnRect.top = editRect.top;
  btnRect.bottom = editRect.bottom;
  btnRect.right = editRect.right;

  //Calculate the width of the button based on the text we are going to put on in
  CDC* pDC{pParent->GetDC()};
  int nButtonWidth{0};
  if (pDC != nullptr)
  {
    //Setup the DC
    CFont* pNewFont{pParent->GetFont()};
    CFont* pOldFont{pDC->SelectObject(pNewFont)};

    //Get the button width
    CSize SizeText{pDC->GetTextExtent(_T("    "), 4)}; //We allow for some space around the button caption
    pDC->LPtoDP(&SizeText);
    nButtonWidth = SizeText.cx;
    SizeText = pDC->GetTextExtent(m_sModifyButtonCaption, m_sModifyButtonCaption.GetLength());
    pDC->LPtoDP(&SizeText);
    nButtonWidth += SizeText.cx;

    //Restore the DC
    pDC->SelectObject(pOldFont);
  }
  else
    nButtonWidth = btnRect.Height()*8/10; //width is 8/10 of height

  //Now that we have the width we can fill out the rest of the buttons rect position
  btnRect.left = btnRect.right - nButtonWidth;

  //Tidy up the device context
  if (pDC != nullptr)
#pragma warning(suppress: 26486)
    pParent->ReleaseDC(pDC);

  //Resize the edit control
  SetWindowPos(nullptr, 0, 0, editRect.Width() - btnRect.Width() - 1, editRect.Height(), SWP_NOACTIVATE | SWP_NOMOVE | SWP_SHOWWINDOW | SWP_NOZORDER);

  //dynamically create the modify button control
  DWORD dwButtonStyles{WS_TABSTOP | WS_VISIBLE | WS_CHILD | WS_GROUP};
  if (GetStyle() & WS_DISABLED)
    dwButtonStyles |= WS_DISABLED;
#pragma warning(suppress: 26486)
  if (!m_ctrlModifyButton.Create(m_sModifyButtonCaption, dwButtonStyles, btnRect, pParent, nModifyButtonID))
    return false;

  //Set the tab order of the button
  m_ctrlModifyButton.SetWindowPos(this, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);

  //tell the button to call this class when it is clicked
  m_ctrlModifyButton.SetBuddy(this);

  //ensure it is using the same font as the parent
  m_ctrlModifyButton.SetFont(pParent->GetFont());

  return true;
}

#pragma warning(suppress: 26434)
void CGetFolderControl::OnEnable(BOOL bEnable)
{
  //Let the base class do its thing
  __super::OnEnable(bEnable);

  //reflect the enabled state of this control to the buddy button control
  m_ctrlModifyButton.EnableWindow(bEnable);
}

#pragma warning(suppress: 26434)
void CGetFolderControl::OnShowWindow(BOOL bShow, UINT nStatus)
{
  //Let the base class do its thing
  __super::OnShowWindow(bShow, nStatus);

  //reflect the visibility state of this control to the buddy button control
  m_ctrlModifyButton.ShowWindow(bShow ? SW_SHOW : SW_HIDE);
}

void CGetFolderControl::Edit()
{
  //Retrieve the current folder to use as the initial value
  CString sDir;
  GetWindowText(sDir);

  //Bring up the common dialog and if the user hits ok, then set the text in this control to the new folder
  CChooseDirDlg dlg{this, m_sCommonDialogTitle};
  if (SUCCEEDED(dlg.DoModal(sDir)))
    SetWindowText(sDir);
}

bool CGetFolderControl::LoadStringResources()
{
  //Use the default strings, a derived class can change this implementation or call the "Set.." functions directly
  if (!m_sModifyButtonCaption.LoadString(IDS_DDXFOLDER_EDIT_TEXT))
    return false;
  if (!m_sModifyButtonToolTipText.LoadString(IDS_DDXFOLDER_TT_MODIFY))
    return false;
  if (!m_sCommonDialogTitle.LoadString(IDS_DDXFOLDER_PLEASE_SPECIFY_A_FOLDER))
    return false;
  return true;
}

CString CGetFolderControl::GetOverwritePromptString(_In_ const CString& sFolder)
{
  CString sReturn;
  AfxFormatString1(sReturn, IDP_DDXFOLDEROVERWRITE_PROMPT, sFolder);
  return sReturn;
}

bool DDX_FolderControl(_In_ CDataExchange* pDX, _In_ int nIDC, _Inout_ CGetFolderControl& rControl, _In_ UINT nModifyButtonID)
{
  //Check to see if the control already exists at the start of this function
  const bool bControlAlreadyExists{rControl.GetSafeHwnd() != nullptr};

  //Let DDX_Control do most of the work
  DDX_Control(pDX, nIDC, rControl);

  //What will be the return value from this method (assume the worst)
  bool bSuccess{false};

  //Setup the other various settings if required
  if (rControl.GetSafeHwnd() != nullptr)
  {
    if (!bControlAlreadyExists)
    {
      rControl.LoadStringResources();
      rControl.InitializeAutoComplete();
      bSuccess = rControl.AddModifyButton(nModifyButtonID);
    }
    else
      bSuccess = true;
  }

  return bSuccess;
}

#pragma warning(suppress: 26461)
void DDX_FolderValue(_In_ CDataExchange* pDX, _In_ CGetFolderControl& rControl, _Inout_ CString& sFolder)
{
  if (pDX->m_bSaveAndValidate)
    rControl.GetWindowText(sFolder);
  else
    rControl.SetWindowText(sFolder);
}

#pragma warning(suppress: 26429)
void DDV_FolderControlIsFolder(_In_ CDataExchange* pDX, _In_ const CGetFolderControl& rControl, _In_ UINT nFailureResourceID, _In_ UINT nHelpID)
{
  if (pDX->m_bSaveAndValidate)
  {
    CString sFolder;
    rControl.GetWindowText(sFolder);

    //Get to see if the item is a folder
    const DWORD dwAttributes{GetFileAttributes(sFolder)};
    if ((dwAttributes != INVALID_FILE_ATTRIBUTES) && ((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0))
    {
      AfxMessageBox(nFailureResourceID, MB_OK | MB_ICONEXCLAMATION, nHelpID);
#pragma warning(suppress: 26486)
      pDX->PrepareEditCtrl(rControl.GetDlgCtrlID());
      pDX->Fail();
    }
  }
}

#pragma warning(suppress: 26429)
void DDV_FolderControlMustExist(_In_ CDataExchange* pDX, _In_ const CGetFolderControl& rControl, _In_ UINT nFailureResourceID, _In_ UINT nHelpID)
{
  if (pDX->m_bSaveAndValidate)
  {
    CString sFolder;
    rControl.GetWindowText(sFolder);
    const DWORD dwAttributes{GetFileAttributes(sFolder)};
    if (dwAttributes == INVALID_FILE_ATTRIBUTES)
    {
      AfxMessageBox(nFailureResourceID, MB_OK | MB_ICONEXCLAMATION, nHelpID);
#pragma warning(suppress: 26486)
      pDX->PrepareEditCtrl(rControl.GetDlgCtrlID());
      pDX->Fail();
    }
  }
}

#pragma warning(suppress: 26429)
void DDV_FolderControlOverwritePrompt(_In_ CDataExchange* pDX, _In_ CGetFolderControl& rControl, _In_ UINT nHelpID)
{
  if (pDX->m_bSaveAndValidate)
  {
    CString sFolder;
    rControl.GetWindowText(sFolder);
    const DWORD dwAttributes{GetFileAttributes(sFolder)};
    if (dwAttributes != INVALID_FILE_ATTRIBUTES)
    {
      //Prompt the user if we can overwrite and if we get a negative indication, then fail the validation
      if (AfxMessageBox(rControl.GetOverwritePromptString(sFolder), MB_YESNO | MB_ICONQUESTION, nHelpID) != IDYES)
      {
#pragma warning(suppress: 26486)
        pDX->PrepareEditCtrl(rControl.GetDlgCtrlID());
        pDX->Fail();
      }
    }
  }
}

#pragma warning(suppress: 26429)
void DDV_FolderControlNotEmpty(_In_ CDataExchange* pDX, _In_ const CGetFolderControl& rControl, _In_ UINT nFailureResourceID, _In_ UINT nHelpID)
{
  if (pDX->m_bSaveAndValidate)
  {
    CString sFolder;
    rControl.GetWindowText(sFolder);
    
    //Check to see if the file exists
    if (sFolder.IsEmpty())
    {
      AfxMessageBox(nFailureResourceID, MB_OK | MB_ICONEXCLAMATION, nHelpID);
#pragma warning(suppress: 26486)
      pDX->PrepareEditCtrl(rControl.GetDlgCtrlID());
      pDX->Fail();
    }
  }
}
