/*
Module : DDXFolder.h
Purpose: Defines the interface for a MFC control to get a directory/folder using the IFileDialog interface

Copyright (c) 1997 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Defines //////////////////////////////////////////////////

#pragma once

#ifndef __DDXFOLDER_H__
#define __DDXFOLDER_H__

#ifndef CDDXFOLDER_EXT_CLASS
#define CDDXFOLDER_EXT_CLASS
#endif //#ifndef CDDXFOLDER_EXT_CLASS

#ifndef CDDXFOLDER_EXT_API
#define CDDXFOLDER_EXT_API
#endif //#ifndef CDDXFOLDER_EXT_API

#ifndef __ATLWIN_H__
#pragma message("DDXFolder as of v1.33 requires ATL support to implement its functionality. If your project is MFC only, then you need to update it to include ATL support")
#endif //#ifndef __ATLWIN_H__


//////////////////// Includes /////////////////////////////////////////////////

#ifndef _SHLOBJ_H_
#pragma message("To avoid this message, please put shlobj.h in your pre compiled header (normally stdafx.h)")
#include <shlobj.h>
#endif //#ifndef _SHLOBJ_H_


//////////////////// Classes //////////////////////////////////////////////////

//forward declaration
class CGetFolderControl;

class CDDXFOLDER_EXT_CLASS CDDXFolderModifyButton : public CButton
{
public:
//Constructors / Destructors
  CDDXFolderModifyButton() noexcept;
  CDDXFolderModifyButton(const CDDXFolderModifyButton&) = delete;
  CDDXFolderModifyButton(CDDXFolderModifyButton&&) = delete;
  ~CDDXFolderModifyButton() = default; //NOLINT(modernize-use-override)

//Methods
  CDDXFolderModifyButton& operator=(const CDDXFolderModifyButton&) = delete;
  CDDXFolderModifyButton& operator=(CDDXFolderModifyButton&&) = delete;
  void SetBuddy(_In_opt_ CGetFolderControl* pBuddy) noexcept;

protected:
//Methods
  BOOL PreTranslateMessage(MSG* pMsg) override;

//Message handlers
  afx_msg void OnClicked();

//Member variables
  CGetFolderControl* m_pBuddy;
  bool m_bFirstPreTranslateMessageCall;
  CToolTipCtrl m_ToolTip;

  DECLARE_MESSAGE_MAP()
};


class CDDXFOLDER_EXT_CLASS CGetFolderControl : public CEdit
{
public:
//Constructors / Destructors
  CGetFolderControl() noexcept;
  CGetFolderControl(const CGetFolderControl&) = delete;
  CGetFolderControl(CGetFolderControl&&) = delete;
  ~CGetFolderControl() = default; //NOLINT(modernize-use-override)

//Methods
  CGetFolderControl& operator=(const CGetFolderControl&) = delete;
  CGetFolderControl& operator=(CGetFolderControl&&) = delete;
  bool SubclassEdit(_In_ HWND hEdit, _In_ UINT nModifyButtonID = -1);
  CDDXFolderModifyButton& GetModifyButton() noexcept { return m_ctrlModifyButton; };

//Accessors/Mutators
  //Set/Get the caption used in the common dialog
  void SetCommonDialogTitle(_In_ const CString& sTitle) { m_sCommonDialogTitle = sTitle; };
  _NODISCARD CString GetCommonDialogTitle() const { return m_sCommonDialogTitle; };

  //Get/Set the Modify button caption 
  void SetModifyButtonCaption(_In_ const CString& sCaption) { m_sModifyButtonCaption = sCaption; };
  _NODISCARD CString GetModifyButtonCaption() const { return m_sModifyButtonCaption; };

  //Get/Set the ToolTip text for the edit button
  void SetModifyButtonToolTipText(_In_ const CString& sText) { m_sModifyButtonToolTipText = sText; };
  _NODISCARD CString GetModifyButtonToolTipText() const { return m_sModifyButtonToolTipText; };

  //Auto completion settings
  void SetAutoComplete(_In_ bool bAutoComplete);
  _NODISCARD bool GetAutoComplete() const noexcept { return m_bAutoComplete; };
  void SetAutoCompleteFlags(_In_ DWORD dwFlags) noexcept { m_dwAutoCompleteFlags = dwFlags; };
  _NODISCARD DWORD GetAutoCompleteFlags() const noexcept { return m_dwAutoCompleteFlags; };

//Misc virtual functions
  virtual CString GetOverwritePromptString(_In_ const CString& sFolder);
  virtual void Edit(); //Bring up the folder picker dialog
  virtual bool AddModifyButton(_In_ UINT nModifyButtonID);
  virtual bool LoadStringResources();
  virtual HRESULT InitializeAutoComplete();

protected:
//Member variables
  CDDXFolderModifyButton m_ctrlModifyButton;
  CString m_sCommonDialogTitle;
  CString m_sModifyButtonCaption;
  CString m_sModifyButtonToolTipText;
  bool m_bAutoComplete;
  DWORD m_dwAutoCompleteFlags;
  ATL::CComPtr<IAutoComplete2> m_pAutoCompleteObject;
  ATL::CComPtr<IACList2> m_pAutoCompleteSourceObject;

//Message handlers
  afx_msg void OnEnable(BOOL bEnable);
  afx_msg void OnShowWindow(BOOL bShow, UINT nStatus);

  DECLARE_MESSAGE_MAP()
};


/////////////// MFC Data exchange routines //////////////////////////

bool CDDXFOLDER_EXT_API DDX_FolderControl(_In_ CDataExchange* pDX, _In_ int nIDC, _Inout_ CGetFolderControl& rControl, _In_ UINT nModifyButton = 0xFFFFFFFF);
void CDDXFOLDER_EXT_API DDX_FolderValue(_In_ CDataExchange* pDX, _In_ CGetFolderControl& rControl, _Inout_ CString& sFolder);
void CDDXFOLDER_EXT_API DDV_FolderControlIsFolder(_In_ CDataExchange* pDX, _In_ const CGetFolderControl& rControl, _In_ UINT nFailureResourceID, _In_ UINT nHelpID = 0xFFFFFFFF);
void CDDXFOLDER_EXT_API DDV_FolderControlMustExist(_In_ CDataExchange* pDX, _In_ const CGetFolderControl& rControl, _In_ UINT nFailureResourceID, _In_ UINT nHelpID = 0xFFFFFFFF);
void CDDXFOLDER_EXT_API DDV_FolderControlOverwritePrompt(_In_ CDataExchange* pDX, _In_ CGetFolderControl& rControl, _In_ UINT nHelpID = 0xFFFFFFFF);
void CDDXFOLDER_EXT_API DDV_FolderControlNotEmpty(_In_ CDataExchange* pDX, _In_ const CGetFolderControl& rControl, _In_ UINT nFailureResourceID, _In_ UINT nHelpID = 0xFFFFFFFF);


#endif //#ifndef __DDXFOLDER_H__
